local _,rematch = ...
local frame = RematchFrame
local incFrame = RematchIncomingFrame

rematch.petsToLoad = {{},{},{}} -- used in loading process, the actual pets/abilities expected at the end of a load
rematch.currentLoadOut = {{},{},{}} -- used in loading process, the current loadout pets/abilities
rematch.loadCandidates = {} -- list of petID candidates that may form the final load

-- returns true if team needs loaded, nil otherwise
function rematch:PetsNeedLoading(team)
	if not team or not rematch.savedTeams[team] then
		return -- no need to load that which does not exist
	end
	rematch:LoadPets(team,true) -- second parameter means we just load petsToLoad
	for i=1,3 do
		for j=1,4 do
			if rematch.currentLoadOut[i][j] ~= rematch.petsToLoad[i][j] then
				return true -- something is different, we do need to load
			end
		end
	end
end

-- takes a 4-entry table (saved[index] or current[index]) and adds it to rematch.loadCandidates if
-- it's an actual pet the user owns and it hasn't been in candidates yet
function rematch:AddToCandidates(petTable)
	local petID = petTable[1]
	local candidates = rematch.loadCandidates
	if petID and petID~=0 and type(petID)=="string" and C_PetJournal.GetPetInfoByPetID(petID) then
		for i=1,#candidates do
			if candidates[i]==petID then
				return -- pet is already a candidate
			end
		end
		tinsert(candidates,petID)
	end
end

-- removes a candidate petID, typically any existing leveling epts when we're about to insert a leveling pet
function rematch:RemoveCandidate(petID)
	local candidates = rematch.loadCandidates
	for i=1,#candidates do
		if candidates[i]==petID then
			tremove(candidates,i)
			return
		end
	end
end

-- go through each ability and confirm it can be used by that pet. if not, mark it 0 to ignore the ability
-- (whatever ability exists when the pet swaps in will be used)
function rematch:ValidateAbilities(petTable)
	local petID = petTable[1]
	local speciesID,_,level = C_PetJournal.GetPetInfoByPetID(petID)
	if not speciesID then
		for i=2,4 do
			petTable[i]=0 -- no idea what pet this is, probably won't get loaded, but ignore abilities anyway
		end
		return
	end
	C_PetJournal.GetPetAbilityList(speciesID,rematch.abilityList,rematch.levelList)
	for i=2,4 do
		local abilityKnown
		for j=1,#rematch.abilityList do
			if petTable[i]==rematch.abilityList[j] and rematch.levelList[j]<=level then
				abilityKnown = true -- ability is not only known but pet has enough levels to use it
				break
			end
		end
		if not abilityKnown then
			petTable[i] = 0
		end
	end
end

-- loads the named team. due to pets maybe missing (speciesID or caged pet) and leveling pets, a
-- team to load may only bear passing resemblance to the team itself. this populates a table
-- rematch.petsToLoad with what pets we want at the end of the loading process.
-- if loadTablesOnly is true, it will stop short of loading the pets, for PetsNeedLoading to use
-- if loadTablesOnly is false, it will then start the loadProcessor which does the actual loading.
function rematch:LoadPets(team,loadTablesOnly)

	if not loadTablesOnly then
		rematch:SummonedPetMayChange()
	end

	local toLoad = rematch.petsToLoad -- 'toLoad' is a table of pets and abilities we will actually load
	local saved = rematch.savedTeams[team] -- 'saved' is the team we want to load
	local current = rematch.currentLoadOut -- 'current' is the current loadout, pets already loaded
	local candidates = rematch.loadCandidates -- list of petIDs that will potentially load

	local levelingPetID = rematch:GetLevelingPetID()
	local levelingSavedSlot -- slot in a saved set that contains a leveling pet
	local levelingCurrentSlot -- slot in current pets that contains a leveling pet

	-- clear tables/load current
	wipe(candidates)
	for i=1,3 do
		wipe(current[i])
		current[i][1],current[i][2],current[i][3],current[i][4] = C_PetJournal.GetPetLoadOutInfo(i)
	end

	-- add saved pets to candidates first
	for i=1,3 do
		if saved[i][1]==0 then
			levelingSavedSlot = i -- if a leveling pet, note its saved slot
		else
			rematch:AddToCandidates(saved[i])
		end
	end
	-- add current pets to candidates afterwards
	for i=1,3 do
		if current[i][1]==levelingPetID then
			current[i][1] = 0
			levelingCurrentSlot = i
		else
			rematch:AddToCandidates(current[i])
		end
	end

	-- at this point we have a list of candidates. now insert the leveling pet if one is used
	if levelingPetID then
		if levelingSavedSlot then -- saved slot has priority
			rematch:RemoveCandidate(levelingPetID) -- remove leveling pet from wherever it is
			tinsert(candidates,levelingSavedSlot,levelingPetID)
		elseif levelingCurrentSlot and rematch.settings.KeepLeveling then -- then current slot
			rematch:RemoveCandidate(levelingPetID)
			tinsert(candidates,levelingCurrentSlot,levelingPetID)
		end
	end

	-- at this point the first three candidates are the actual pets to load huzzah

	-- now populate petsToLoad with the first three pets's abilities
	for i=1,3 do
		rematch:FillLoadAbilities(candidates[i],toLoad[i],team)
		rematch:ValidateAbilities(toLoad[i]) -- make sure abilities can be loaded, mark 0 if not
	end

	if loadTablesOnly then
		return -- we only wanted to see exactly what would load, can leave now
	end

	rematch.loadProcessor = rematch.loadProcessor or CreateFrame("Frame")
	rematch.loadProcessor.startedLoading = GetTime()
	rematch.loadProcessor:SetScript("OnUpdate",rematch.ProcessLoad)

end

-- candidates are just petIDs, this function finds the abilities that should go with this petID
-- if they're not known, abilities of 0 mean "whatever is loaded"
function rematch:FillLoadAbilities(petID,petTable,team)
	local saved = rematch.savedTeams[team] -- 'saved' is the team we want to load
	local current = rematch.currentLoadOut -- 'current' is the current loadout, pets already loaded
	wipe(petTable)
	petTable[1] = petID
	for i=1,3 do
		if saved[i][1]==petID then
			for j=2,4 do
				petTable[j] = saved[i][j]
			end
			return
		end
	end
	for i=1,3 do
		if current[i][1]==petID then
			for j=2,4 do
				petTable[j] = current[i][j]
			end
			return
		end
	end
	-- if we made it here, the intended petID isn't in a saved team or a current pet, 0 its abilities
	for j=2,4 do
		petTable[j] = 0
	end
end

-- this is the meat of the load process. every frame, it will go through the current
-- loadout pets and see if any pets or abilities need swapped in. if so, it will load
-- the pet or ability and immediately leave. once no swaps happen, this stops
function rematch:ProcessLoad(elapsed)

	local info = rematch.info
	local toLoad = rematch.petsToLoad

  -- if 5 seconds haven't elapsed
	if GetTime()<(self.startedLoading+5) then
		-- go through each pet and see if it's already loaded. if not, load it and come back next frame
		for i=1,3 do
			wipe(info)
			info[1],info[2],info[3],info[4] = C_PetJournal.GetPetLoadOutInfo(i)
			if toLoad[i][1] and info[1]~=toLoad[i][1] then
				C_PetJournal.SetPetLoadOutInfo(i,toLoad[i][1])
				return
			elseif toLoad[i][1] then
				-- go through each ability and see if it's already loaded. if not, load it and come back next frame
				for j=2,4 do
					if toLoad[i][j]~=0 and info[j]~=toLoad[i][j] then
						C_PetJournal.SetAbility(i,j-1,toLoad[i][j])
						return
					end
				end
			end
		end
	end

	-- if we made it this far, pets are all loaded! (or we gave up after timeout)
	if frame:IsVisible() then
		for i=1,3 do
			frame.current[i].shine:SetCooldown(0,1) -- shine the current pets
		end
	end
	if PetJournal then
		PetJournal_UpdatePetLoadOut() -- update petjournal if it's open
	end
	self:SetScript("OnUpdate",nil)

	if rematch.settings.AutoShow and rematch.selectedTeam and rematch.targetedName==rematch.selectedTeam and not rematch.settings.LockWindow then
		frame:Hide()
	end

end
